/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

 An implementation of the Boyer-Moore-Horspool algorithm.
 See R. N. Horspool: Practical fast searching in strings -
 "Software: Practice and Experience" Vol. 10 Pagg. 501-506.

 $Id: bmh.c,v 1.1 1999/05/14 09:51:55 loreti Exp $

 *------------------------------------------------------*/

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>
#endif /* TEST */

#include <string.h>

static void BMH_init(char *);
static int  BMH_match(char *, char *);

static size_t   bmPL;           /* Pattern length */
static int      bmPos[256];     /* Shift table */

#define A_LEN   (sizeof(bmPos) / sizeof(bmPos[0]))

#ifdef TEST
int main()
{

/**
 | A simple test program; exit with a backslash as first character
 | for "string".  No check for input buffer overflow.
**/

  char  string[128], pat[128];
  int   match, offset, count;
  char *start;

  for (;;) {
    printf("String ? ");
    fgets(string, 128, stdin);
    if (string[0] == '\\') exit(EXIT_SUCCESS);
    string[strlen(string) - 1] = '\0';

    printf("Pattern ? ");
    fgets(pat, 128, stdin);
    pat[strlen(pat) - 1] = '\0';

    printf("String: \"%s\"; pattern: \"%s\"\n", string, pat);

/**
 | Finds ALL the occurrences of "pat" in "string": "count" is their
 | total number; "offset" the starting position in string; "match"
 | the position (relative to string[offset]) of the next occurrence,
 | or 0.
**/

    BMH_init(pat);
    offset = count = 0;
    start  = string;
    while ((match = BMH_match(pat, start)) >= 0) {
      ++count;
      offset += match;
      printf("Match at string[%d]\n", offset);
      ++offset;
      start = string + offset;
    }
    printf("Total: %d match%s\n", count, (count == 1 ? "" : "es"));
  }
}
#endif /* TEST */

static void BMH_init(
  char *pattern
) {

/**
 | Initialisation of the Boyer-Moore-Horspool shift table
**/

  int i;

  bmPL = strlen(pattern);
  for (i=0; i<A_LEN; i++)           bmPos[i]                 = bmPL;
  for (i=1; i<bmPL; i++, pattern++) bmPos[(size_t) *pattern] = bmPL - i;
}

static int BMH_match(
  char *pat,
  char *text
) {
  size_t i, j, k, l_text;

  i = j = bmPL;
  l_text = strlen(text);

  while (j > 0   &&   i <= l_text) {
    k = i;
    while (j > 0   &&   pat[j-1] == text[k-1]) {
      k--;
      j--;
    }
    if (j > 0) {
      i += bmPos[(size_t) text[i-1]];
      j  = bmPL;
    } else {
      return (int) k;
    }
  }
  return -1;
}
