/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

 An implementation of a simplified Boyer-Moore algorithm
 with a single table.

 $Id: boyermoore.c,v 1.1 1999/05/14 09:51:55 loreti Exp $

 *------------------------------------------------------*/

#include <string.h>

static void bmInit(char *);
static int bmSearch(char *, char *);
static int bmMatch(char *, char *);

static size_t bmPL;
static short bmPos[256];
static int pos;

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>

int main()
{

/**
 | A simple test program; exit with a backslash as first character
 | for "string".  No check for input buffer overflow.
**/

  int  i;
  char string[128], pat[128];

  for (;;) {
    printf("String ? ");
    fgets(string, 128, stdin);
    if (string[0] == '\\') exit(EXIT_SUCCESS);
    string[strlen(string) - 1] = '\0';

    printf("Pattern ? ");
    fgets(pat, 128, stdin);
    pat[strlen(pat) - 1] = '\0';

    printf("String: \"%s\"; pattern: \"%s\"\n", string, pat);

/**
 | Finds ALL the occurrences of "pat" in "string"
**/

    bmInit(pat);
    while ((i = bmMatch(pat, string)) >= 0) {
      printf("Match at string[%d]\n", i);
    }
    printf("NMatches: %d\n", bmSearch(pat, string));
  }
}
#endif /* TEST */

static void bmInit(
  char *pattern
) {
  int i;

  bmPL = strlen(pattern);
  memset(bmPos, 0, sizeof(bmPos));
  for (i=0; i<bmPL; i++) bmPos[(size_t) pattern[i]] = i;
  pos = -1;
}

static int bmSearch(
  char *pat,
  char *text
){
  int nMatch = 0;

  while (bmMatch(pat, text) >= 0) nMatch++;
  return nMatch;
}

static int bmMatch(
  char *pat,
  char *text
){
  int lastPos;

  pos++;
  lastPos = strlen(text) - bmPL;
  while (pos <= lastPos) {
    int step;
    int j = bmPL - 1;

    while (j >= 0   &&   text[pos + j]  ==  pat[j])    j--;
    if (j < 0) return pos;
    step = j - bmPos[(size_t) text[pos + j]];
    if (step > 1) pos += step;
        else      pos++;
  }
  pos = -1;
  return -1;
}
