//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  ++39(49) 827-7216     FAX: ++39(49) 827-7102
//     EMail:  loreti@padova.infn.it
//     WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
//     An implementation of the Boyer-Moore string search
//     algorithm; see boyermoore.cxx for more comments.
//
//     $Id: boyermoore.h,v 1.2 1999/06/28 05:25:03 loreti Exp $
//
// -------------------------------------------------------------

#ifndef BOYERMOORE_H_
#define BOYERMOORE_H_

#include <stdexcept>

class bmPattern {
    friend class bmMatcher;

  public:
    bmPattern (const char *patt, const int ignoreCase) throw(std::bad_alloc);
    bmPattern (const bmPattern &rhs) throw(std::bad_alloc);
    ~bmPattern () throw();
    bmPattern & operator = (const bmPattern &rhs) throw(std::bad_alloc);

  private:
    static const int BM_TABLE_LEN;
    int              bmCase;
    size_t           bmPatLen;
    short           *bmPos;
    char            *bmPat;
};

// bmMatcher: the copy constructor and the destructor are declared
//            private and never defined to prevent their use.

class bmMatcher {
  public:
    bmMatcher     (const bmPattern &, const char *) throw(std::bad_alloc);
    ~bmMatcher    () throw();
    int nMatches  () throw();
    int nextMatch () throw();

  private:
    const bmPattern &pat;
    int              pos;
    size_t           lastPos;
    char            *text;

    bmMatcher              (const bmMatcher &);
    bmMatcher & operator = (const bmMatcher &);
};

#endif // BOYERMOORE_H_
